@@ -1,11 +1,19 @@
 // Copyright (C) 2023 The Qt Company Ltd.
 // SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only
 
+#ifdef USE_LINEGRAPH
 #include <QtGraphs/qlineseries.h>
+#endif
+#ifdef USE_SCATTERGRAPH
 #include <QtGraphs/qscatterseries.h>
+#endif
+#ifdef USE_SPLINEGRAPH
 #include <QtGraphs/qsplineseries.h>
-#include <private/pointrenderer_p.h>
+#endif
+#include <QtQuick/private/qquickdraghandler_p.h>
+#include <QtQuick/private/qquicktaphandler_p.h>
 #include <private/axisrenderer_p.h>
+#include <private/pointrenderer_p.h>
 #include <private/qabstractseries_p.h>
 #include <private/qgraphsview_p.h>
 #include <private/qxyseries_p.h>
@@ -19,6 +27,7 @@ static const char *TAG_POINT_SELECTED_COLOR = "pointSelectedColor";
 static const char *TAG_POINT_SELECTED = "pointSelected";
 static const char *TAG_POINT_VALUE_X = "pointValueX";
 static const char *TAG_POINT_VALUE_Y = "pointValueY";
+static const char *TAG_POINT_INDEX = "pointIndex";
 
 PointRenderer::PointRenderer(QGraphsView *graph)
     : QQuickItem(graph)
@@ -47,6 +56,14 @@ PointRenderer::PointRenderer(QGraphsView *graph)
     )QML").arg(QString::number((int) defaultSize()));
     m_tempMarker = new QQmlComponent(qmlEngine(m_graph), this);
     m_tempMarker->setData(qmlData.toUtf8(), QUrl());
+
+    m_tapHandler = new QQuickTapHandler(this);
+    connect(m_tapHandler, &QQuickTapHandler::singleTapped,
+            this, &PointRenderer::onSingleTapped);
+    connect(m_tapHandler, &QQuickTapHandler::doubleTapped,
+            this, &PointRenderer::onDoubleTapped);
+    connect(m_tapHandler, &QQuickTapHandler::pressedChanged,
+            this, &PointRenderer::onPressedChanged);
 }
 
 PointRenderer::~PointRenderer()
@@ -58,10 +75,17 @@ qreal PointRenderer::defaultSize(QXYSeries *series)
 {
     qreal size = 16.0;
     if (series != nullptr) {
+#ifdef USE_LINEGRAPH
         if (auto line = qobject_cast<QLineSeries *>(series))
             size = qMax(size, line->width());
-        else if (auto spline = qobject_cast<QSplineSeries *>(series))
+#endif
+#if defined(USE_LINEGRAPH) && defined(USE_SPLINEGRAPH)
+        else
+#endif
+#ifdef USE_SPLINEGRAPH
+            if (auto spline = qobject_cast<QSplineSeries *>(series))
             size = qMax(size, spline->width());
+#endif
     }
     return size;
 }
@@ -94,38 +118,57 @@ void PointRenderer::reverseRenderCoordinates(
              / (-1 * m_areaHeight * flipY * m_maxVertical);
 }
 
-void PointRenderer::updatePointDelegate(
-    QXYSeries *series, PointGroup *group, qsizetype pointIndex, qreal x, qreal y)
+PointRenderer::SeriesStyle PointRenderer::getSeriesStyle(PointGroup *group)
 {
     auto theme = m_graph->theme();
-    auto marker = group->markers[pointIndex];
-    auto &rect = group->rects[pointIndex];
 
     const auto &seriesColors = theme->seriesColors();
     const auto &borderColors = theme->borderColors();
+
     qsizetype index = group->colorIndex % seriesColors.size();
-    QColor color = series->color().alpha() != 0 ? series->color() : seriesColors.at(index);
+    QColor color = group->series->color().alpha() != 0 ? group->series->color() : seriesColors.at(index);
+
+    QColor selectedColor = group->series->selectedColor().alpha() != 0
+                               ? group->series->selectedColor()
+                               : m_graph->theme()->singleHighlightColor();
+
     index = group->colorIndex % borderColors.size();
     QColor borderColor = borderColors.at(index);
     qreal borderWidth = theme->borderWidth();
-    QColor selectedColor = series->selectedColor().alpha() != 0
-                               ? series->selectedColor()
-                               : m_graph->theme()->singleHighlightColor();
+
+    return {
+        color,
+        selectedColor,
+        borderColor,
+        borderWidth
+    };
+}
+
+void PointRenderer::updatePointDelegate(
+    QXYSeries *series, PointGroup *group, qsizetype pointIndex, qreal x, qreal y)
+{
+    const auto style = getSeriesStyle(group);
+
+    auto marker = group->markers[pointIndex];
+    auto &rect = group->rects[pointIndex];
+
     if (marker->property(TAG_POINT_SELECTED).isValid())
         marker->setProperty(TAG_POINT_SELECTED, series->isPointSelected(pointIndex));
     if (marker->property(TAG_POINT_COLOR).isValid())
-        marker->setProperty(TAG_POINT_COLOR, color);
+        marker->setProperty(TAG_POINT_COLOR, style.color);
     if (marker->property(TAG_POINT_BORDER_COLOR).isValid())
-        marker->setProperty(TAG_POINT_BORDER_COLOR, borderColor);
+        marker->setProperty(TAG_POINT_BORDER_COLOR, style.borderColor);
     if (marker->property(TAG_POINT_BORDER_WIDTH).isValid())
-        marker->setProperty(TAG_POINT_BORDER_WIDTH, borderWidth);
+        marker->setProperty(TAG_POINT_BORDER_WIDTH, style.borderWidth);
     if (marker->property(TAG_POINT_SELECTED_COLOR).isValid())
-        marker->setProperty(TAG_POINT_SELECTED_COLOR, selectedColor);
+        marker->setProperty(TAG_POINT_SELECTED_COLOR, style.selectedColor);
     const auto point = series->points().at(pointIndex);
     if (marker->property(TAG_POINT_VALUE_X).isValid())
         marker->setProperty(TAG_POINT_VALUE_X, point.x());
     if (marker->property(TAG_POINT_VALUE_Y).isValid())
         marker->setProperty(TAG_POINT_VALUE_Y, point.y());
+    if (marker->property(TAG_POINT_INDEX).isValid())
+        marker->setProperty(TAG_POINT_INDEX, pointIndex);
 
     marker->setX(x - marker->width() / 2.0);
     marker->setY(y - marker->height() / 2.0);
@@ -155,15 +198,101 @@ void PointRenderer::updateLegendData(QXYSeries *series, QLegendData &legendData)
     series->d_func()->setLegendData(legendDataList);
 }
 
+void PointRenderer::onSingleTapped(QEventPoint eventPoint, Qt::MouseButton button)
+{
+    Q_UNUSED(button)
+
+    for (auto &&group : m_groups) {
+        if (!group->series->isVisible())
+            continue;
+
+        if (!group->series->isSelectable() && !group->series->isDraggable())
+            continue;
+
+        int index = 0;
+        for (auto &&rect : group->rects) {
+            if (rect.contains(eventPoint.position())) {
+                emit group->series->clicked(group->series->at(index).toPoint());
+                return;
+            }
+            index++;
+        }
+    }
+}
+
+void PointRenderer::onDoubleTapped(QEventPoint eventPoint, Qt::MouseButton button)
+{
+    Q_UNUSED(button)
+
+    for (auto &&group : m_groups) {
+        if (!group->series->isVisible())
+            continue;
+
+        if (!group->series->isSelectable() && !group->series->isDraggable())
+            continue;
+
+        int index = 0;
+        for (auto &&rect : group->rects) {
+            if (rect.contains(eventPoint.position())) {
+                emit group->series->doubleClicked(group->series->at(index).toPoint());
+                return;
+            }
+            index++;
+        }
+    }
+}
+
+void PointRenderer::onPressedChanged()
+{
+    if (m_tapHandler->isPressed()) {
+        for (auto &&group : m_groups) {
+            if (!group->series->isVisible())
+                continue;
+
+            if (!group->series->isSelectable() && !group->series->isDraggable())
+                continue;
+
+            int index = 0;
+            for (auto &&rect : group->rects) {
+                if (rect.contains(m_tapHandler->point().position())) {
+                    m_pressedGroup = group;
+                    m_pressedPointIndex = index;
+                    emit group->series->pressed(m_pressedGroup->series->at(index).toPoint());
+                }
+                index++;
+            }
+        }
+    } else {
+        if (m_pressedGroup
+            && m_pressedGroup->series->isSelectable()
+            && m_pressedGroup->series->isVisible()) {
+            if (m_pressedGroup->rects[m_pressedPointIndex].contains(
+                    m_tapHandler->point().position())) {
+                if (m_pressedGroup->series->isPointSelected(m_pressedPointIndex))
+                    m_pressedGroup->series->deselectPoint(m_pressedPointIndex);
+                else
+                    m_pressedGroup->series->selectPoint(m_pressedPointIndex);
+                m_previousDelta = QPoint(0, 0);
+                emit m_pressedGroup->series->released(
+                    m_pressedGroup->series->at(m_pressedPointIndex).toPoint());
+            }
+        }
+    }
+}
+
+#ifdef USE_SCATTERGRAPH
 void PointRenderer::updateScatterSeries(QScatterSeries *series, QLegendData &legendData)
 {
+    auto group = m_groups.value(series);
+    const auto style = getSeriesStyle(group);
+
     if (series->isVisible()) {
-        auto group = m_groups.value(series);
         auto &&points = series->points();
         group->rects.resize(points.size());
         for (int i = 0; i < points.size(); ++i) {
             qreal x, y;
             calculateRenderCoordinates(m_graph->m_axisRenderer, points[i].x(), points[i].y(), &x, &y);
+            y *= series->valuesMultiplier();
             if (group->currentMarker) {
                 updatePointDelegate(series, group, i, x, y);
             } else {
@@ -175,21 +304,18 @@ void PointRenderer::updateScatterSeries(QScatterSeries *series, QLegendData &leg
     } else {
         hidePointDelegates(series);
     }
-    // TODO: When fill color is added to the scatterseries use it instead for
-    // the color. QTBUG-122434
-    legendData = {series->color(), series->color(), series->name()};
-}
 
+    legendData = { style.color, style.borderColor, series->name() };
+}
+#endif
+
+#ifdef USE_LINEGRAPH
 void PointRenderer::updateLineSeries(QLineSeries *series, QLegendData &legendData)
 {
-    auto theme = m_graph->theme();
     auto group = m_groups.value(series);
+    const auto style = getSeriesStyle(group);
 
-    const auto &seriesColors = theme->seriesColors();
-    qsizetype index = group->colorIndex % seriesColors.size();
-    QColor color = series->color().alpha() != 0 ? series->color() : seriesColors.at(index);
-
-    group->shapePath->setStrokeColor(color);
+    group->shapePath->setStrokeColor(style.color);
     group->shapePath->setStrokeWidth(series->width());
     group->shapePath->setFillColor(QColorConstants::Transparent);
 
@@ -210,6 +336,7 @@ void PointRenderer::updateLineSeries(QLineSeries *series, QLegendData &legendDat
         for (int i = 0; i < points.size(); ++i) {
             qreal x, y;
             calculateRenderCoordinates(m_graph->m_axisRenderer, points[i].x(), points[i].y(), &x, &y);
+            y *= series->valuesMultiplier();
             if (i == 0)
                 painterPath.moveTo(x, y);
             else
@@ -227,19 +354,17 @@ void PointRenderer::updateLineSeries(QLineSeries *series, QLegendData &legendDat
         hidePointDelegates(series);
     }
     group->shapePath->setPath(painterPath);
-    legendData = {color, color, series->name()};
+    legendData = { style.color, style.borderColor, series->name() };
 }
+#endif
 
+#ifdef USE_SPLINEGRAPH
 void PointRenderer::updateSplineSeries(QSplineSeries *series, QLegendData &legendData)
 {
-    auto theme = m_graph->theme();
     auto group = m_groups.value(series);
+    const auto style = getSeriesStyle(group);
 
-    const auto &seriesColors = theme->seriesColors();
-    qsizetype index = group->colorIndex % seriesColors.size();
-    QColor color = series->color().alpha() != 0 ? series->color() : seriesColors.at(index);
-
-    group->shapePath->setStrokeColor(color);
+    group->shapePath->setStrokeColor(style.color);
     group->shapePath->setStrokeWidth(series->width());
     group->shapePath->setFillColor(QColorConstants::Transparent);
 
@@ -263,6 +388,8 @@ void PointRenderer::updateSplineSeries(QSplineSeries *series, QLegendData &legen
             qreal x, y;
             calculateRenderCoordinates(m_graph->m_axisRenderer, points[i].x(), points[i].y(), &x, &y);
 
+            qreal valuesMultiplier = series->valuesMultiplier();
+            y *= valuesMultiplier;
             if (i == 0) {
                 painterPath.moveTo(x, y);
             } else {
@@ -277,6 +404,9 @@ void PointRenderer::updateSplineSeries(QSplineSeries *series, QLegendData &legen
                                            fittedPoints[j].y(),
                                            &x2,
                                            &y2);
+
+                y1 *= valuesMultiplier;
+                y2 *= valuesMultiplier;
                 painterPath.cubicTo(x1, y1, x2, y2, x, y);
                 ++j;
             }
@@ -294,8 +424,9 @@ void PointRenderer::updateSplineSeries(QSplineSeries *series, QLegendData &legen
     }
 
     group->shapePath->setPath(painterPath);
-    legendData = {color, color, series->name()};
+    legendData = { style.color, style.borderColor, series->name() };
 }
+#endif
 
 void PointRenderer::handlePolish(QXYSeries *series)
 {
@@ -389,7 +520,50 @@ void PointRenderer::handlePolish(QXYSeries *series)
                     group->currentMarker->create(group->currentMarker->creationContext()));
                 item->setParent(this);
                 item->setParentItem(this);
+                QQuickDragHandler *handler = new QQuickDragHandler(item);
+                handler->setEnabled(series->isDraggable());
+                connect(series, &QXYSeries::draggableChanged, this, [handler, series]() {
+                    handler->setEnabled(series->isDraggable());
+                });
                 group->markers << item;
+                group->dragHandlers << handler;
+
+                connect(handler, &QQuickDragHandler::translationChanged, this, [&]() {
+                    if (m_pressedGroup) {
+                        float w = width();
+                        float h = height();
+                        double maxVertical
+                            = m_graph->m_axisRenderer->m_axisVerticalValueRange > 0
+                                  ? 1.0 / m_graph->m_axisRenderer->m_axisVerticalValueRange
+                                  : 100.0;
+                        double maxHorizontal
+                            = m_graph->m_axisRenderer->m_axisHorizontalValueRange > 0
+                                  ? 1.0 / m_graph->m_axisRenderer->m_axisHorizontalValueRange
+                                  : 100.0;
+
+                        QPoint currentDelta =
+                            m_pressedGroup->dragHandlers.at(m_pressedPointIndex)
+                                ->activeTranslation().toPoint();
+                        QPoint delta = currentDelta - m_previousDelta;
+                        m_previousDelta = currentDelta;
+
+                        qreal deltaX = delta.x() / w / maxHorizontal;
+                        qreal deltaY = -delta.y() / h / maxVertical;
+
+                        QPointF point = m_pressedGroup->series->at(m_pressedPointIndex)
+                                        + QPointF(deltaX, deltaY);
+                        m_pressedGroup->series->replace(m_pressedPointIndex, point);
+                    }
+                });
+                connect(handler, &QQuickDragHandler::grabChanged, this,
+                        [&](QPointingDevice::GrabTransition transition, QEventPoint point) {
+                            Q_UNUSED(point)
+
+                            if (transition == QPointingDevice::UngrabExclusive ||
+                                transition == QPointingDevice::UngrabPassive) {
+                                m_previousDelta = QPoint(0, 0);
+                            }
+                        });
             }
         } else if (markerCount > pointCount) {
             for (qsizetype i = pointCount; i < markerCount; ++i)
@@ -408,12 +582,24 @@ void PointRenderer::handlePolish(QXYSeries *series)
     }
 
     QLegendData legendData;
+#ifdef USE_SCATTERGRAPH
     if (auto scatter = qobject_cast<QScatterSeries *>(series))
         updateScatterSeries(scatter, legendData);
-    else if (auto line = qobject_cast<QLineSeries *>(series))
+#endif
+#if defined(USE_SCATTERGRAPH) && defined(USE_LINEGRAPH)
+    else
+#endif
+#ifdef USE_LINEGRAPH
+        if (auto line = qobject_cast<QLineSeries *>(series))
         updateLineSeries(line, legendData);
-    else if (auto spline = qobject_cast<QSplineSeries *>(series))
+#endif
+#if defined(USE_LINEGRAPH) && defined(USE_SPLINEGRAPH)
+    else
+#endif
+#ifdef USE_SPLINEGRAPH
+        if (auto spline = qobject_cast<QSplineSeries *>(series))
         updateSplineSeries(spline, legendData);
+#endif
 
     updateLegendData(series, legendData);
 }
@@ -450,81 +636,6 @@ void PointRenderer::afterUpdate(QList<QAbstractSeries *> &cleanupSeries)
     Q_UNUSED(cleanupSeries);
 }
 
-bool PointRenderer::handleMouseMove(QMouseEvent *event)
-{
-    if (!m_pressedGroup || !m_pressedGroup->series->isVisible())
-        return false;
-
-    if (m_pointPressed && m_pressedGroup->series->isDraggable()) {
-        float w = width();
-        float h = height();
-        double maxVertical = m_graph->m_axisRenderer->m_axisVerticalValueRange > 0
-                                 ? 1.0 / m_graph->m_axisRenderer->m_axisVerticalValueRange
-                                 : 100.0;
-        double maxHorizontal = m_graph->m_axisRenderer->m_axisHorizontalValueRange > 0
-                                   ? 1.0 / m_graph->m_axisRenderer->m_axisHorizontalValueRange
-                                   : 100.0;
-
-        QPoint delta = m_pressStart - event->pos();
-
-        qreal deltaX = -delta.x() / w / maxHorizontal;
-        qreal deltaY = delta.y() / h / maxVertical;
-
-        QPointF point = m_pressedGroup->series->at(m_pressedPointIndex) + QPointF(deltaX, deltaY);
-        m_pressedGroup->series->replace(m_pressedPointIndex, point);
-
-        m_pressStart = event->pos();
-        m_pointDragging = true;
-
-        return true;
-    }
-    return false;
-}
-
-bool PointRenderer::handleMousePress(QMouseEvent *event)
-{
-    bool handled = false;
-    for (auto &&group : m_groups) {
-        if (!group->series->isVisible())
-            continue;
-
-        if (!group->series->isSelectable() && !group->series->isDraggable())
-            continue;
-
-        int index = 0;
-        for (auto &&rect : group->rects) {
-            if (rect.contains(event->pos())) {
-                m_pointPressed = true;
-                m_pressStart = event->pos();
-                m_pressedGroup = group;
-                m_pressedPointIndex = index;
-                handled = true;
-            }
-            index++;
-        }
-    }
-    return handled;
-}
-
-bool PointRenderer::handleMouseRelease(QMouseEvent *event)
-{
-    bool handled = false;
-    if (!m_pointDragging && m_pointPressed && m_pressedGroup
-        && m_pressedGroup->series->isSelectable() && m_pressedGroup->series->isVisible()) {
-        if (m_pressedGroup->rects[m_pressedPointIndex].contains(event->pos())) {
-            if (m_pressedGroup->series->isPointSelected(m_pressedPointIndex)) {
-                m_pressedGroup->series->deselectPoint(m_pressedPointIndex);
-            } else {
-                m_pressedGroup->series->selectPoint(m_pressedPointIndex);
-            }
-            handled = true;
-        }
-    }
-    m_pointPressed = false;
-    m_pointDragging = false;
-    return handled;
-}
-
 bool PointRenderer::handleHoverMove(QHoverEvent *event)
 {
     bool handled = false;
